import json
import os
import sys

from .module_manager import ModuleManager
from .utils.exp_util import get_relative_imports
from .utils.git_util import parse_url_from_git
from ...mate_config import MateConfig
from .sources.remote import RemoteDataSource
from .sources.local.local import LocalDataSource
from pipreqs import pipreqs

from .package import Package
from typing import Optional
import ipdb


class ModuleRepository:
    def __init__(self, config: MateConfig, run_local_api_server: bool = False):
        self.config = config
        self.package_manager = ModuleManager(config)
        self.remote = RemoteDataSource()
        self.local = LocalDataSource(config)
        self.__generate_pip_requirements(self.config.project)

    @staticmethod
    def init_project(project_name: str):

        if not os.path.exists(project_name):
            os.mkdir(project_name)
            os.chdir(project_name)
        else:
            print("Project already exists")
            sys.exit(1)

        mate_json = os.path.join("mate.json")
        if not os.path.exists(mate_json):
            dic = {
                "project": project_name,
            }
            # create mate.json

            with open(mate_json, "w") as f:
                json.dump(dic, f, indent=4)
        else:
            print("Project already exists")
            sys.exit(1)

        if not os.path.exists(project_name):
            os.mkdir(project_name)
            init__file = os.path.join(project_name, "__init__.py")
            open(init__file, "a").close()
        try:
            folders = ["experiments", "models", "data", "trainers"]
            for folder in folders:
                os.makedirs(os.path.join(project_name, folder), exist_ok=True)
                init__file = os.path.join(project_name, folder, "__init__.py")
                if not os.path.exists(init__file):
                    open(init__file, "a").close()
            print(
                "Project {} created, run `cd {}` to enter the project folder".format(
                    project_name, project_name
                )
            )

        except Exception as e:
            print(e)

    def install_url(self, url: str, *args, **kwargs):
        self.package_manager.install_package(url, *args, **kwargs)

    def auto(self, command: str, *args):
        if command == "export":
            self.__generate_sub_pip_reqs()
        elif command in ["init", "fix", "i"]:
            self.__generate__init__(self.config.project)

    def __generate__init__(self, root: str = None):
        init__py = os.path.join(root, "__init__.py")
        if not os.path.exists(init__py):
            with open(init__py, "w") as f:
                f.write("")
            print(f"Created {init__py}")

        for folder in os.listdir(root):
            path = os.path.join(root, folder)
            if not os.path.isdir(path) or path == "__pycache__" or "." in folder:
                continue
            init__py = os.path.join(path, "__init__.py")
            if not os.path.exists(init__py):
                with open(init__py, "w") as f:
                    f.write("")
                print(f"Created {init__py}")
            self.__generate__init__(path)

    def __generate_sub_pip_reqs(self):

        for folder in os.listdir(self.config.project):
            path = os.path.join(self.config.project, folder)
            if os.path.isdir(path) and "__" not in folder:
                for subfolder in os.listdir(path):
                    subpath = os.path.join(path, subfolder)
                    if os.path.isdir(subpath) and "__" not in subfolder:
                        self.__generate_pip_requirements(subpath)
                # self.__generate_pip_requirements(path)

    def __generate_mate_dependencies(self, path):
        # ipdb.set_trace()

        files = [f for f in os.listdir(path) if f.endswith(".py") and "__" not in f]
        relative_imports = [get_relative_imports(os.path.join(path, f)) for f in files]
        # flatten array to unique set
        relative_imports = set(
            [item for sublist in relative_imports for item in sublist]
        )

        url_git = parse_url_from_git()

        if url_git is None:
            print("No git url found, skipping dependencies.json")
            return

        deps = set()
        for module in relative_imports:
            url = self.config.project + "/" + module.replace(".", "/")
            if url_git:
                url = url_git + url
            deps.add(url)

        if len(deps) == 0:
            return

        with open(os.path.join(path, "dependencies.json"), "w") as f:
            deps = {"dependencies": list(deps)}
            json.dump(deps, f, indent=4)
            print(f"Generated dependencies.json for {path}")

    def __generate_pip_requirements(self, path):

        imports = pipreqs.get_all_imports(path)
        # import_info_remote = pipreqs.get_imports_info(imports)
        import_info_local = pipreqs.get_import_local(imports)

        if "experiments" in path:
            self.__generate_mate_dependencies(path)

        import_info = []

        if path == self.config.project:
            pipreqs.generate_requirements_file(
                "requirements.txt", import_info_local, "=="
            )
        else:
            pipreqs.generate_requirements_file(
                os.path.join(path, "requirements.txt"), import_info_local, "=="
            )
            print(f"Generated requirements.txt for {path}")

        for im in import_info_local:

            name = im["name"]
            version = im["version"]

            res = {
                "name": name,
                "version": version,
            }

            import_info.append(res)

        return {"pip": import_info}

    def list(self, module: str = None):
        if module == None:
            return self.local.summary()
        return self.local.list(module)

    def get_mate_summary(self):
        return self.local.summary()

    def install_package(self, package: Package):
        self.local.install_package(package)



from .utils.bunch import Bunch
from .mate import Mate
import inspect
import ipdb
import sys
from pydoc import locate
from typing import Any


def parse_signature(class_name, method_name: str):
    return tuple(
        val
        for (name, val) in inspect.signature(
            getattr(class_name, method_name)
        ).parameters.items()
        if name != "self"
    )


def get_methods_with_arguments(class_name):
    return {
        k: parse_signature(class_name, k)
        for k in class_name.__dict__.keys()
        if not k.startswith("_")
    }

    


def prettify_method(method_name, annotation, in_depth: bool = False):
    def cleanup(ann):
        return str(ann).replace("<class ", "").replace(">", "").replace("'", "")

    def pretty(m):
        return ",\n\t".join(
            [
                f"{m.name}: {cleanup(m.annotation)}"
                + (f"={m.default}" if m.default != inspect._empty else "")
                for m in annotation
            ]
        )

    source_code = inspect.getsource(getattr(Mate, method_name))
    description = (
        source_code.split('"""')[1] if '"""' in source_code and in_depth else ""
    )
    return f" {method_name}\n\t{pretty(annotation)}\n" + description


def print_help(methods):
    for method_name, annotation in methods.items():
        print(prettify_method(method_name, annotation) + "\n")


def convert_str_to_data(input):
    try:
        return int(input)
    except ValueError:
        try:
            return float(input)
        except ValueError:
            if input in ["True", "true"]:
                return True
            elif input in ["False", "false"]:
                return False

    return input


def parse_run_params(args: list):
    params = {}
    for arg in args:
        key, value = arg.split("=")
        params[key[2:]] = convert_str_to_data(value)
    return params


def main():
    methods = get_methods_with_arguments(Mate)
    args = sys.argv[1:]
    raw_method_args = args[1:]
    actions = tuple(method.replace("_", "-") for method in methods) + (
        "--help",
        "-h",
    )
    if len(args) == 0 or not args[0] in actions or args[0] in ("--help", "-h"):
        print_help(methods)
    else:
        assert args[0] in actions, print_help(methods)
        action = args[0].replace("-", "_")
        if len(args) > 1 and args[1] in ("--help", "-h"):
            print(prettify_method(action, methods[args[0]], in_depth=True))
        else:
            method_args_types = tuple(param.annotation for param in methods[action])
            method_args_defaults = tuple(
                tuple(param.default for param in methods[action])
            )
            method_args = tuple(
                (
                    method_type(raw_method_arg)
                    if (raw_method_arg is not None)
                    else method_default
                )
                for (method_type, method_default, raw_method_arg) in zip(
                    method_args_types, method_args_defaults, raw_method_args
                )
            )

            if action == "init":
                Mate.init(*method_args)
            else:
                mate = Mate()
                method_args_len = len(method_args)
                hparams_len = len(raw_method_args) - method_args_len
                if hparams_len > 0:
                    run_params = parse_run_params(args[method_args_len + 1 :])
                    mate.run_params = run_params
                function = getattr(mate, action)
                # # add Nones for missing method_args
                # method_args += (None,) * (
                #     len(inspect.signature(function).parameters) - 1
                # )
                function(*method_args)



import os
from yerbamate.api.data.sources.local import io
import json
from yerbamate.api.mate_api import MateAPI

import sys

import ipdb
from .utils import utils


from typing import Optional

from .mate_config import MateConfig



class Mate:


    def __init__(self, init=False):
        self.root_folder = ""
        self.save_path = ""
        self.current_folder = os.path.dirname(__file__)
        self.config: Optional[MateConfig] = None
        self.__findroot()
        self.api = MateAPI(self.config)
        self.run_params = None

 
    @staticmethod
    def init(project_name: str):
        MateAPI.init_project(project_name)
 

    def export(self):
        self.api.auto("export")

    def list(
        self,
        module_name: str = None,
        output_json: bool = True,
    ):
        li = self.api.list(module_name)
        if output_json:
            print(json.dumps(li, indent=4))
        else:
            print(li)


    def auto(self, command: str):
        """
        auto init -> creates __init__.py recursively
        auto export -> creates requirements.txt for export
        """
        self.api.auto(command)


    def summary(self, output_json: bool = True):
        print(json.dumps(self.api.summary(), indent=4))

    def clone(self, source_model: str, target_model: str):
        pass

    def snapshot(self, model_name: str):
        pass

    def train(self, model: str, exp: str , *args, **kwargs):
        # ipdb.set_trace()
        module = [self.config.project, "experiments", model, exp]
        module = ".".join(module)

        try:
            __import__(module)
        except ModuleNotFoundError:
            print(f"Experiment {model}/{exp} not found")
            print("Available experiments:")
            self.list("experiments")
            sys.exit(1)



    def install(self, url: str, *args, **kwargs):
        """
        Adds a dependency to a model.
        """
        self.api.install_url(url, *args, **kwargs)

    def exec(self, model: str, params: str, exec_file: str):
        params = "parameters" if params == "" or params == "None" else params
        print(f"Executing model {model} with result of: {params}")
        _, model, _ = self.__get_trainer(model, params)

        self.__load_exec_function(exec_file)(model)

    # def board(self):

    #     self.api.start_mateboard()

    # def __list_packages(self, folder: str):
    #     return io.list_packages(self.root_folder, folder)

    # def __update_mate_version(self):
    #     utils.migrate_mate_version(self.config, self.root_folder)

    def __findroot(self):
        """
        Method in charge of finding the root folder of the project and reading the content of mate.json
        """
        self.root_folder, self.config = io.find_root()
        self.root_save_folder = self.config.results_folder

    # def __get_trainer(self, params: str):
    #     if self.trainer is None:

    #         conf, save_path = self.package_manager.load_experiment(params)

    #         print(conf)

    #         map_key_value = {
    #             "save_path": self.save_path,
    #             "save_dir": self.save_path,
    #         }
    #         root_module = f"{self.root_folder}"

    #         self.trainer = Trainer.create(conf, root_module, map_key_value)

    #     return self.trainer

    # def __fit(self, params: str):

    #     trainer = self.__get_trainer(params)

    #     self.__parse_and_validate_params(params)

    #     if self.is_restart:
    #         checkpoint_path = os.path.join(self.save_path, "checkpoints", "last.ckpt")
    #         trainer.fit(ckpt_path=checkpoint_path)
    #     else:
    #         trainer.fit()



import hashlib
import json
import os
import re
from shutil import copytree, rmtree
import validators
from yerbamate.mate_config import MateConfig

from .utils.gitdir import download
import ipdb


class ModuleManager:
    root_path: str = ""

    def __init__(self, conf: MateConfig):
        self.conf = conf

        self.init_package_cache()

    def init_package_cache(self):

        folder_name = ".mate"
        if not os.path.exists(folder_name):
            os.mkdir(folder_name)

        self.root_path = os.path.join(os.getcwd(), folder_name)

    def get_path(self, *args):
        return os.path.join(self.root_path, *args)


    def check_auto_install_reqs(self, *args, **kwargs):
        if len(args) < 0:
            return False
        if args[0] == "-y":
            return True

    def auto_install_manager(self, *args, **kwargs):
        if self.check_auto_install_reqs(*args, **kwargs):
            if len(args) > 1:
                pm = args[1]
                if pm in ["pip", "conda"]:
                    return pm
            else:
                input = input("Install with pip or conda? [pip/conda]: ")
                if input in ["pip", "conda"]:
                    return input

            while input not in ["pip", "conda"]:
                input = input("Install with pip or conda? [pip/conda]: ")
                if input in ["pip", "conda"]:
                    return input
            
            return None
        return None

    def install_package(self, url: str, *args, **kwargs):

        if url.count("/") > 3 and "github.com" not in url:

            url = url.split("/")
            # oalee/deep-vision/deepnet/models/resnet to https://github.com/oalee/deep-vision/tree/main/deepnet/models/resnet
            url = f"https://github.com/{url[0]}/{url[1]}/tree/main/{'/'.join(url[2:])}"

        assert validators.url(url), "Invalid url"

        package_install_dst = self.__install_package(url)

        # ipdb.set_trace()

        # check for requirements.txt
        if package_install_dst:
            requirements_path = os.path.join(package_install_dst, "requirements.txt")
            if os.path.exists(requirements_path):

                if self.check_auto_install_reqs(*args, **kwargs):
                    pm = self.auto_install_manager(*args, **kwargs)
                    if pm == "pip":
                        os.system(f"pip install -r {requirements_path}")
                    elif pm == "conda":
                        os.system(f"conda install --file {requirements_path}")
                    else:
                        print("Skipping requirements installation")
                    
                else:
                    cmd = input(
                        "Requirements found. Do you want to install them with pip? [y/conda/n]: "
                    )
                    if cmd == "y":
                        os.system(f"pip install -r {requirements_path}")
                    elif cmd == "conda":
                        os.system(f"conda install --file {requirements_path}")
                    else:
                        print("Skipping requirements installation")
            else:
                print("No requirements found. Manually check and install dependencies.")
            # read dependencies.json
            dependencies_path = os.path.join(package_install_dst, "dependencies.json")
            if os.path.exists(dependencies_path):
                with open(dependencies_path, "r") as f:
                    dependencies = json.load(f)
                for dependency in dependencies["dependencies"]:
                    # ipdb.set_trace()
                    self.install_package(dependency, *args, **kwargs)

        # update history

        print(f"Module installed at {package_install_dst}")

    def __install_package(self, url):

        # just git url
        base_git_url = "/".join(url.split("/")[:7])

        url_hash = hashlib.sha1(base_git_url.encode("utf-8")).hexdigest()

        output_path = self.get_path(url_hash)
        output_dir = output_path

        if not os.path.exists(output_path):
            os.makedirs(output_path, exist_ok=True)

        root_module = url.split("/")[-2]
        module_name = url.split("/")[-1]

        if root_module in ["models", "data", "experiments", "trainers"]:
            dest_path = os.path.join(
                os.getcwd(), self.conf.project, root_module, module_name
            )

        else:
            print("Could not automatically determine the module type. Please specify.")
            command = input(f"Is {root_module}/{module_name} Correct [y,n]?")
            if command == "y":
                dest_path = os.path.join(
                    os.getcwd(), self.conf.project, root_module, module_name
                )
            else:
                dest_path = input(
                    "Please specify the correct path: (e.g. models.my_model)"
                )
                while dest_path.count(".") < 0:
                    dest_path = input(
                        "Please specify the correct path: (e.g. models.my_model) "
                    )
                dest_path = dest_path.split(".")
                dest_path = os.path.join(os.getcwd(), self.conf.project, *dest_path)

        while os.path.exists(dest_path):
            cmd = input(
                "Package already exists, options: [o]verwrite, [c]ancel, [r]ename: "
            )

            if cmd == "o":
                rmtree(dest_path)
                break

            elif cmd == "c":
                return

            elif cmd == "r":
                dest_path = input(
                    "Please specify the correct path: (e.g. models.my_model) "
                )
                while dest_path.count(".") < 0:
                    dest_path = input(
                        "Please specify the correct path: (e.g. models.my_model) "
                    )
                    print(dest_path.count("."))
                dest_path = dest_path.split(".")
                dest_path = os.path.join(os.getcwd(), self.conf.project, *dest_path)
                break

        print(f"Downloading {url}")

        download(url, output_dir=output_dir)

        src_module = url.split("tree")[1].split("/")[2:]
        src_module = os.path.join(output_dir, *src_module)

        copytree(src_module, dest_path)
        rmtree(output_dir)
        return dest_path

